"""Classes and methods for working with the physical Mebo robot"""
import logging
import socket
import sys
import time
from abc import ABC
from collections import namedtuple
from functools import partial
from ipaddress import (
    AddressValueError,
    IPv4Network,
    IPv4Address
)

from xml.etree.ElementTree import fromstring as xmlfromstring

from requests import Session
from requests.exceptions import (
    ConnectionError,
    HTTPError
)

from zeroconf import ServiceBrowser, Zeroconf

from .exceptions import (
    MeboCommandError,
    MeboDiscoveryError,
    MeboRequestError,
    MeboConnectionError,
    MeboConfigurationError
)

from mebo.stream.session import (
    RTSPSession,
)


logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

Broadcast = namedtuple('Broadcast', ['ip', 'port', 'data'])
WirelessNetwork = namedtuple('WirelessNetwork', ['ssid', 'mac', 'a', 'q', 'si', 'nl', 'ch'])

NORTH = 'n'
NORTH_EAST = 'ne'
EAST = 'e'
SOUTH_EAST = 'se'
SOUTH = 's'
SOUTH_WEST = 'sw'
WEST = 'w'
NORTH_WEST = 'nw'
DIRECTIONS = {NORTH, NORTH_EAST, EAST, SOUTH_EAST, SOUTH, SOUTH_WEST, WEST, NORTH_WEST}


class _MeboMDNSListener:

    def remove_service(self, zeroconf, type, name):
        logging.debug("Service %s removed", name)

    def add_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)
        logging.debug("Service %s added, service info: %s", name, info)


class ComponentFactory:
    """Factory class for generating classes of components"""

    @classmethod
    def _from_parent(cls, name, **actions):
        """Generates a class of the given type as a subclass of component

        :param name: Name of the generated class
        :type name: str
        :param actions: action names-> `callables`, which are closures using the parents shared request infrastructure
        :type actions: dict
        """
        cls = type(name, (Component,), actions)
        cls.__doc__ = f"""{name.upper()} Component"""
        return cls(actions=actions.keys())


class Component(ABC):
    """Abstract base class for all robot components"""

    def __init__(self, actions):
        self.actions = actions

    def __repr__(self):
        return '<{} actions={}>'.format(self.__class__, self.actions)


class Mebo:
    """Mebo represents a single physical robot"""

    # port used by mebo to broadcast its presence
    BROADCAST_PORT = 51110

    # port used to establish media (RTSP) sessions
    RTSP_PORT = 6667

    def __init__(self, ip=None, auto_connect=False):
        """Initializes a Mebo robot object and establishes an http connection

        If no ip or network is supplied, then we will autodiscover the mebo using mDNS.

        :param ip: IPv4 address of the robot as a string.
        :type ip: str
        :param auto_connect: if True, will autodiscover and connect at object creation time
        :type auto_connect: bool

        >>> m = Mebo()
        >>> m.connect()
        >>> m2 = Mebo(auto_connect=True)
        """
        self._session = Session()
        self._ip = None

        if ip:
            self.ip = ip
        elif auto_connect:
            self.connect()

        self._arm = None
        self._wrist = None
        self._claw = None
        self._speaker = None

        self._rtsp_session = None

    @property
    def endpoint(self):
        return 'http://{}'.format(self.ip)

    @property
    def ip(self):
        """The IP of the robot on the LAN

        This value is either provided explicitly at creation time or autodiscovered via mDNS
        """
        if self._ip is None:
            raise MeboConfigurationError('No configured or discovered value for ip address')
        return self._ip

    @ip.setter
    def ip(self, value):
        try:
            addr = IPv4Address(value)
            self._ip = addr
        except AddressValueError:
            raise MeboConfigurationError(f'Value {addr} set for IP is invalid IPv4 Address')

    @property
    def media(self):
        """an rtsp session representing the media streams (audio and video) for the robot"""
        if self._rtsp_session is None:
            url = f'rtsp://{self.ip}/streamhd/'
            self._rtsp_session = RTSPSession(
                url,
                port=self.RTSP_PORT,
                username='stream',
                realm='realm',
                user_agent='python-mebo'
            )
        return self._rtsp_session

    # TODO: rip this out, or change it to a hearbeat
    # listener which gets established, this has nothing to do with discovery
    # instead, use mDNS
    def _get_broadcast(self, address, timeout=10):
        """ Attempts to receive the UDP broadcast signal from Mebo
        on the supplied address. Raises an exception if no data is received
        before 'timeout' seconds.

        :param address: the broadcast address to bind
        :param timeout: how long the socket should wait without receiving data before raising socket.timeout

        :returns: A Broadcast object that containing the source IP, port, and data received.
        :raises: `socket.timeout` if no data is received before `timeout` seconds
        """
        logging.debug(f"reading from: {address}:{Mebo.BROADCAST_PORT}")
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
            s.bind((address, Mebo.BROADCAST_PORT))
            s.settimeout(timeout)
            data, source = s.recvfrom(4096)
            logging.debug(f"Data received: {data}:{source}")
            return Broadcast(source[0], source[1], data)

    def _setup_video_stream(self):
        self._request(req='feedback_channel_init')
        self._request(req='set_video_gop', value=40)
        self._request(req='set_date', value=time.time())
        self._request(req='set_video_gop', value=40, speed=1)
        self._request(req='set_video_gop', value=40, speed=2)
        self._request(req='set_video_bitrate', value=600)
        self._request(req='set_video_bitrate', value=600, speed=1)
        self._request(req='set_video_bitrate', value=600, speed=2)
        self._request(req='set_resolution', value='720p')
        self._request(req='set_resolution', value='720p', speed=1)
        self._request(req='set_resolution', value='720p', speed=2)
        self._request(req='set_video_qp', value=42)
        self._request(req='set_video_qp', value=42, speed=1)
        self._request(req='set_video_qp', value=42, speed=2)
        self._request(req='set_video_framerate', value=20)
        self._request(req='set_video_framerate', value=20, speed=1)
        self._request(req='set_video_framerate', value=20, speed=2)

    def _get_stream(self, address, timeout=10):
        pass

    def _get_mdns(self, key="_camera._tcp.local."):
        try:
            zeroconf = Zeroconf()
            listener = _MeboMDNSListener()
            browser = ServiceBrowser(zeroconf, key, listener)
            time.sleep(1)
            for name, record in browser.services.items():
                info = zeroconf.get_service_info(record.key, record.alias)
                # note: the library we're using keeps these keys and values as bytes
                return info.properties[b'ip'].decode('ascii')
        finally:
            zeroconf.close()

    def _discover(self):
        """
        Runs the discovery scan to find Mebo on your LAN

        :returns: The IP address of the Mebo found on the LAN
        :raises: `MeboDiscoveryError` if discovery times out or the API probe produces a 40X or 50x status code
        """
        try:
            logging.debug('Looking for Mebo...')
            ip = self._get_mdns()
            return ip
        except socket.timeout:
            raise MeboDiscoveryError(('Unable to locate Mebo on the network.\n'
                                      '\tMake sure it is powered on and connected to LAN.\n'
                                      '\tIt may be necessary to power cycle the Mebo.'))

    def connect(self):
        """Connect to the mebo control server over HTTP

        If no IP exists for the robot already, the IP will be autodiscovered via mDNS. When there is already an IP, that will be used to make a canary request to get the command server software version. If the robot has been previously connected, no request is made at all.

        :raises: :class:`mebo.exceptions.MeboDiscoveryError` when a mDNS discovery fails
        :raises: :class:`mebo.exceptions.MeboConnectionError` when a TCP ConnectionError or HTTPError occurs
        """
        if self._ip is None:
            self.ip = self._discover()
            logging.debug(f'Mebo found at {self.ip}')
        try:
            version = self.version
            logging.debug(f'Mebo {version} connected')
        except (ConnectionError, HTTPError) as e:
            raise MeboConnectionError(f'Unable to connect to mebo: {e}')

    def _request(self, **params):
        """ private function to submit HTTP requests to Mebo's API

        :param params: arguments to pass as query params to the Mebo API. Might
        also include `need_response`, a kwarg dictating whether the caller
        requires the response object from the API.

        :returns: The `requests.HTTPResponse` object if `need_response` is True, `None` otherwise.
        """
        try:
            need_response = params.pop('need_response')
        except KeyError:
            # by default, don't return a response
            need_response = False
        try:
            response = self._session.get(self.endpoint, params=params)
            response.raise_for_status()
            if need_response:
                return response
        except (ConnectionError, HTTPError) as e:
            raise MeboRequestError(f'Request to Mebo failed: {e}')

    def visible_networks(self):
        """
        Retrieves list of wireless networks visible to Mebo.

        :returns: A dictionary of name to `WirelessNetwork`

        >>> m = Mebo(auto_connect=True)
        >>> print(m.visible_networks())
        """
        resp = self._request(req='get_rt_list', need_response=True)
        et = xmlfromstring(f'{resp.text}')
        visible = {}
        for nw in et.findall('w'):
            visible[nw.find('s').text.strip('"')] = WirelessNetwork(*(i.text.strip('"') for i in nw.getchildren()))
        return visible

    def add_router(self, auth_type, ssid, password, index=1):
        """
        Save a wireless network to the Mebo's list of routers
        """
        self._request(req='setup_wireless_save', auth=auth_type, ssid=ssid, key=password, index=index)

    def set_scan_timer(self, value=30):
        self._request(req='set_scan_timer', value=value)

    def restart(self):
        self._request(req='restart_system')

    def set_timer_state(self, value=0):
        self._request(req='set_timer_state', value=value)

    def get_wifi_cert(self):
        resp = self._request(req='get_wifi_cert', need_response=True)
        _, cert_type = resp.text.split(':')
        return cert_type.strip()

    def get_boundary_position(self):
        """ Gets boundary positions for 4 axes:

        Arm: s_up, s_down
        Claw: c_open, c_close
        Wrist Rotation: w_left & w_right
        Wrist Elevation: h_up, h_down

        :returns: dictionary of functions to boundary positions
        """
        resp = self._request(req='get_boundary_position', need_response=True)
        _, key_value_string = [s.strip() for s in resp.text.split(':')]
        return dict((k, int(v)) for k, v in [ks.strip().split('=') for ks in key_value_string.split('&')])

    @property
    def version(self):
        """returns the software version of the robot

        >>> m = Mebo(auto_connect=True)
        >>> m.version == '03.02.37'
        """
        if not hasattr(self, '_version') or self._version is None: 
            resp = self._request(req='get_version', need_response=True)
            _, version = resp.text.split(':')
            self._version = version.strip()
        return self._version

    def move(self, direction, speed=255, dur=1000):
        """Move the robot in a given direction at a speed for a given duration

        :param direction: map direction to move. 'n', 'ne', 'nw', etc.
        :param speed: a value in the range [0, 255]. default: 255
        :param dur: number of milliseconds the wheels should spin. default: 1000
        """
        directions = {
            NORTH: 'move_forward',
            NORTH_EAST: 'move_forward_right',
            EAST: 'move_right',
            SOUTH_EAST: 'move_backward_right',
            SOUTH: 'move_backward',
            SOUTH_WEST: 'move_backward_left',
            WEST: 'move_left',
            NORTH_WEST: 'move_forward_left'
        }
        direction = directions.get(direction.lower())
        if direction is None:
            raise MeboCommandError(
                'Direction must be one of the map directions: {}'.format(directions.keys()
            ))
        speed = min(speed, 255)
        # there is also a ts keyword that could be passed here.
        self._request(req=direction, dur=dur, value=speed)

    def turn(self, direction):
        """Turns a very small amount in the given direction

        :param direction: one of R or L
        """
        direction = direction.lower()[0]
        if direction not in {"r", "l"}:
            raise MeboCommandError('Direction for turn must be either "right", "left", "l", or "r"')
        call = 'inch_right' if direction == 'r' else 'inch_left'
        self._request(req=call)

    def stop(self):
        self._request(req='fb_stop')

    @property
    def claw(self):
        """ The claw component at the end of Mebo's arm

        >>> m = Mebo(auto_connect=True)
        >>> m.claw.open(dur=1000, **params)
        >>> m.claw.close(dur=400, **params)
        >>> m.claw.stop(**params)
        """
        if self._claw is None:
            claw = ComponentFactory._from_parent(
                'Claw',
                open=partial(self._request, req='c_open'),
                close=partial(self._request, req='c_close'),
                stop=partial(self._request, req='c_stop')
            )
            self._claw = claw
        return self._claw

    @property
    def wrist(self):
        """The wrist component of the robot

        The wrist component has the following actions:
        * rotate clockwise (to the right from the robot's perspective) OR counter-clockwise (to the left from the robot's perspective)
        * raise or lower

        >>> m = Mebo()
        >>> m.wrist.rotate_right(**params)
        >>> m.wrist.rotate_left(**params)
        >>> m.wrist.inch_right(**params)
        >>> m.wrist.inch_left(**params)
        >>> m.wrist.rotate_stop()
        >>> m.wrist.up()
        >>> m.wrist.down()
        >>> m.wrist.lift_stop()
        """
        if self._wrist is None:
            wrist = ComponentFactory._from_parent(
                'Wrist',
                rotate_right=partial(self._request, req='w_right'),
                inch_right=partial(self._request, req='inch_w_right'),
                rotate_left=partial(self._request, req='w_left'),
                inch_left=partial(self._request, req='inch_w_left'),
                rotate_stop=partial(self._request, req='w_stop'),
                up=partial(self._request, req='h_up'),
                down=partial(self._request, req='h_down'),
                lift_stop=partial(self._request, req='h_stop'),
            )
            self._wrist = wrist
        return self._wrist

    @property
    def arm(self):
        """The arm component of mebo

        >>> m = Mebo(auto_connect=True)
        >>> m.arm.up(dur=1000, **params)
        >>> m.arm.down(dur=1000, **params)
        >>> m.arm.stop(**params)
        """
        up = partial(self._request, req='s_up')
        up.__doc__ = """Move the arm up

        :param dur: The duration of the arm movement
        :type dur: int
        """
        down = partial(self._request, req='s_down')
        down.__doc__ = """Move the arm down

        :param dur: The duration of the arm movement
        :type dur: int
        """
        stop = partial(self._request, req='s_stop')
        stop.__doc__ = """Stop the arm"""

        if self._arm is None:
            arm = ComponentFactory._from_parent(
                'Arm',
                up=up,
                down=down,
                stop=stop
            )
            self._arm = arm
        return self._arm

    @property
    def speaker(self):
        """
        >>> m = Mebo(auto_connect=True)
        >>> m.speaker.set_volume(value=6)
        >>> m.speaker.get_volume()
        >>> m.speaker.play_sound(**params)
        """
        if self._speaker is None:
            speaker = ComponentFactory._from_parent(
                'Speaker',
                set_volume=partial(self._request, req='set_spk_volume'),
                play_sound=partial(self._request, req='audio_out0'))
            self._speaker = speaker
        return self._speaker



import logging
import socket
import random
import re

from mebo.stream.decode import (
    RTPPacket,
    RTPDecodeError
)


logger = logging.getLogger(__name__)

MAX_TRIES = 3

SERVER_PORTS = re.compile(r'server_port=(?P<server_rtp>\d{4,5})-(?P<server_rtcp>\d{4,5})')


class RTPStream:

    TIMEOUT = 0
    START_BYTES = bytes([0xfe, 0xed, 0xfa, 0xce])
    RTCP_HEADER = [0x80, 0xc9, 0x0, 0x1]

    @staticmethod
    def choose_port(start=50000, end=60000):
        return random.randrange(start, end, 2)

    def __init__(self, sdp, host=None, timeout=15):
        self.sdp = sdp
        self.host = host or self.sdp.host_info
        self.name = self.sdp.control

        self._media = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._media.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_port = None

        self._rtcp = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._rtcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_rtcp = None

        self._media.settimeout(timeout)
        self._rtcp.settimeout(timeout)

        tries = 0
        while tries < MAX_TRIES:
            try:
                self._media.bind(('', RTPStream.choose_port()))
                self._rtcp.bind(('', self.rtcp_port))
                break
            except Exception as e:
                logger.error('Failed to bind: %s', e)
                tries += 1
        else:
            raise Exception('Unable to bind port')

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ':'.join([self.host, self.name])

    def setup(self, transport_header):
        match = SERVER_PORTS.search(transport_header)
        if not match:
            raise Exception('No server_ports found!')
        self.server_port, self.server_rtcp = map(int, match.groups())
        logger.debug('Server port for RTP on %s: %d', self.name, self.server_port)
        logger.debug('Server port for RTCP on %s: %d', self.name, self.server_rtcp)
        assert self._media.sendto(self.START_BYTES, (self.host, self.server_port))

    @property
    def transport(self):
        # could eventually get unicast info from stream itself maybe
        return f'{self.sdp.profile};unicast;client_port={self.media_port}-{self.rtcp_port}'

    @property
    def media_port(self):
        return self._media.getsockname()[1]

    @property
    def rtcp_port(self):
        return self.media_port + 1

    def capture(self, filename, seconds=None, bytes=None, packets=1000):
        capture_count = 0
        with open(filename, 'wb') as f:
            for i in range(packets):
                raw_packet = self._media.recv(4096)
                rtp = RTPPacket(raw_packet)
                try:
                    decoded = rtp.decode(self.sdp.name)
                    f.write(decoded)
                except RTPDecodeError as e:
                    logger.error('Unable to decode packet: %s', e)
                    continue
                capture_count += 1
        logger.debug('Captured %d packets in %s', capture_count, filename)
        return capture_count



"""
Based on the work here:

    https://gist.github.com/jn0/8b98652f9fb8f8d7afbf4915f63f6726

    http://stackoverflow.com/questions/28022432/receiving-rtp-packets-after-rtsp-setup
    A demo python code that ..

    1) Connects to an IP cam with RTSP
    2) Draws RTP/NAL/H264 packets from the camera
    3) Writes them to a file that can be read with any stock video player (say, mplayer, vlc & other ffmpeg based video-players)

    Done for educative/demonstrative purposes, not for efficiency..!

    written 2015 by Sampsa Riikonen.


Adapted to python3
"""

import struct
import logging

logger = logging.getLogger(__name__)


def get_bits(d, size=32):
    m, M = 0, 2**size - 1
    if not (m <= d <= M):
        raise ValueError(f'{d} must be in the range [{m},{M}]')
    binary = f'{d:0>{size}b}'
    assert len(binary) == size
    return {i: digit for i, digit in enumerate(binary)}


class NALType:

    RESERVED = {0, 30, 31}
    UNIT = set(range(1, 24))

    STAP_A = 24
    STAP_B = 25
    MTAP16 = 26
    MTAP24 = 27
    FU_A = 28
    FU_B = 29

    @classmethod
    def supported(cls):
        return {
            cls.STAP_A,
            cls.FU_A
        }.union(cls.UNIT)

    @classmethod
    def unsupported(cls):
        return {
            cls.STAP_B,
            cls.MTAP16,
            cls.MTAP24,
            cls.FU_B,
        }.union(cls.RESERVED)


class RTPDecodeError(Exception):
    pass


class RTPPacket:

    def __init__(self, packet):
        self.raw = memoryview(packet)
        self.byte_offset = 0
        self.version = 0
        self.padding = 0
        self.contributors = 0
        self.marker = 0
        self.payload_type = 0
        self.sequence_number = 0
        self.timestamp = 0

        # optional extensions
        self.header_id = None
        self.header_length = None
        self.header_contents = None

    @property
    def bytes(self):
        return self.raw.tobytes()

    def parse_header(self):
        # The first 8-bytes represent the RTP header value
        # the header format can be found from:
        # https://en.wikipedia.org/wiki/Real-time_Transport_Protocol
        # https://tools.ietf.org/html/rfc3550#section-5.1
        #
        #     0                   1                   2                   3
        #     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     |V=2|P|X|  CC   |M|     PT      |       sequence number         |
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     |                           timestamp                           |
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #     |           synchronization source (SSRC) identifier            |
        #     +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
        #     |            contributing source (CSRC) identifiers             |
        #     |                             ....                              |
        #     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #
        #     The first twelve octets (96 bits) are present in every RTP packet, while the
        #     list of CSRC identifiers is present only when inserted by a mixer.
        #     The fields have the following meaning:

        #     Required fields:
        #     version (V): 2 bits
        #     padding (P): 1 bit
        #     extension (X): 1 bit
        #     CSRC count (CC): 4 bits
        #     marker (M): 1 bit
        #     payload type (PT): 7 bits
        #     sequence number (M): 16 bits
        #     timestamp: 32 bits
        #     SSRC: 32 bits

        # the first 12 bytes we'll pack into three, 32-bit unsigned ints
        # then use bit shifting/twiddling to get the parts we want
        # consumes the first 12 bytes or 96 bits
        standard_header = self.raw[:12]
        self.byte_offset += 12

        meta, self.timestamp, self.ssrc = struct.unpack('!III', standard_header)
        # first 16 bits of the first uint32
        first_16 = meta >> 16
        self.sequence_number = meta & ((1 << 16) - 1)  # last 16 bits of the first uint32
        logger.debug('Sequence number: %d, Timestamp: %d, SSRC: %x', self.sequence_number, self.timestamp, self.ssrc)

        self.version = first_16 >> 14  # first two bits, 23 == 0x17 == 3 << 14 ==
        if self.version != 2:
            raise RTPDecodeError(f'Unknown version {self.version} error!')
        self.padding = (first_16 >> 13) & 1  # third bit
        self.extension = (first_16 >> 12) & 1  # fourth bit
        self.contributors = first_16 & ((1 << 4) - 1)   # bits 5-8
        self.marker = first_16 & (1 << 7)  # bit 9

        self.payload_type = first_16 & ((1 << 7) - 1)  # last 7 bits of first two bytes

        if self.contributors:
            orig_offset = self.byte_offset
            self.byte_offset += (4 * self.contributors)
            struct_mask = ''.join(['!', 'I' * self.contributors])
            self.contributing_sources = struct.unpack(struct_mask, self.raw[orig_offset: self.byte_offset])
            logger.debug("CRSC identifiers: %s", [f'{csrc:x}' for csrc in self.contributing_sources])

        if self.extension:
            # these are unsigned 16-bit integers 'H' in struct unpack land
            orig_offset = self.byte_offset
            self.byte_offset += 2
            self.header_id = struct.unpack('!H', self.raw[orig_offset: self.byte_offset])

            orig_offset = self.byte_offset
            self.byte_offset += 2
            self.header_length = struct.unpack('!H', self.raw[orig_offset: self.byte_offset])
            logger.debug('Extended header id (%d) with length %d', self.header_id, self.header_length)

            orig_offset = self.byte_offset
            self.byte_offset += (4 * self.header_length)
            self.header_contents = self.raw[orig_offset: self.byte_offset]


    def decode(self, media_type):
        self.parse_header()
        if media_type == 'video':
            return self._decode_video()
        elif media_type == 'audio':
            return self._decode_audio()
        else:
            raise ValueError(f'Unknown sdp media type {media_type}')

    def _decode_audio(self, code='g711a'):
        """ Writes out everything after the header """
        return self.raw[self.byte_offset:]

    def _decode_video(self, codec='h264', packetization_mode=1):
        """ This routine takes a UDP packet, i.e. a string of bytes and ..
        (a) strips off the RTP header
        (b) adds NAL "stamps" to the packets, so that they are recognized as NAL's
        (c) Concantenates frames
        (d) Returns a packet that can be written to disk as such and that is recognized by stock media players as h264 stream
        """
        # this is the sequence of four bytes that identifies a NAL packet.. must be in front of every NAL packet.
        startbytes = b'\x00\x00\x00\x01'

        # OK, all of the header has been consumed.
        # now we enter the NAL packet, as described here:
        # https://tools.ietf.org/html/rfc6184#section-1.3
        # Some quotes from that document:
        """
        5.3. NAL Unit Header Usage


        The structure and semantics of the NAL unit header were introduced in
        Section 1.3.  For convenience, the format of the NAL unit header is
        reprinted below:

            +---------------+
            |0|1|2|3|4|5|6|7|
            +-+-+-+-+-+-+-+-+
            |F|NRI|  Type   |
            +---------------+

        This section specifies the semantics of F and NRI according to this
        specification.

        """
        """
        Table 3.  Summary of allowed NAL unit types for each packetization
                      mode (yes  =  allowed, no  =  disallowed, ig  =  ignore)

            Payload Packet    Single NAL    Non-Interleaved    Interleaved
            Type    Type      Unit Mode           Mode             Mode
            -------------------------------------------------------------
            0      reserved      ig               ig               ig
            1-23   NAL unit     yes              yes               no
            24     STAP-A        no              yes               no
            25     STAP-B        no               no              yes
            26     MTAP16        no               no              yes
            27     MTAP24        no               no              yes
            28     FU-A          no              yes              yes
            29     FU-B          no               no              yes
            30-31  reserved      ig               ig               ig
        """
        # This was also very usefull:
        # http://stackoverflow.com/questions/7665217/how-to-process-raw-udp-packets-so-that-they-can-be-decoded-by-a-decoder-filter-i
        # A quote from that:
        """
        First byte:  [ 3 NAL UNIT BITS | 5 FRAGMENT TYPE BITS]
        Second byte: [ START BIT | END BIT | RESERVED BIT | 5 NAL UNIT BITS]
        Other bytes: [... VIDEO FRAGMENT DATA...]
        """

        self.NAL = self.raw[self.byte_offset]
        self.f = (self.NAL & 0x80) >> 7  # bit 1, forbidden bit. Should be 0
        self.nri = (self.NAL & 0x60) >> 5  # bit 2-3, NAL ref IDC
        self.fragment_type = self.NAL & 0x1F  # ((1 << 5) - 1)
        self.nlu0 = self.f | self.nri

        logger.debug('F:%d, NRI:%d, Type:%d', self.f, self.nri, self.fragment_type)
        # logger.debug('NLU: %d', self.nlu0)

        if self.fragment_type in NALType.UNIT:
            # this means we have either an SPS or a PPS packet
            # they have the meta-info about resolution, etc.
            # more reading for example here:
            # http://www.cardinalpeak.com/blog/the-h-264-sequence-parameter-set/
            if self.fragment_type == 1:
                logger.debug('<<<<< NAL unit type (%d), decoding', self.fragment_type)
            if self.fragment_type == 7:
                logger.debug(">>>>> SPS packet")
            else:
                logger.debug(">>>>> PPS packet")
            # .. notice here that we include the NAL starting sequence "startbytes" and the "First byte"
            return startbytes + self.raw[self.byte_offset:]

        else:
            self.byte_offset += 1

        # let's go to "Second byte"
        # ********* WE ARE AT THE "Second byte" ************
        # The "Type" here is most likely 28, i.e. "FU-A"
        second_byte = self.raw[self.byte_offset]
        self.start = (second_byte & 0x80) >> 7  # 128
        self.end = (second_byte & 0x40) >> 6  # 64
        self.reserved = (second_byte & 0x20) >> 5  # 32
        self.nlu1 = second_byte & 0x1F

        if self.start:  # OK, this is a first fragment in a movie frame
            logger.debug('>>>>> first fragment found')
            self.nlu = bytes([self.nlu0 | self.nlu1])  # Create "[3 NAL UNIT BITS | 5 NAL UNIT BITS]"
            # possibly decode bytes as latin-1?
            head = startbytes + self.nlu   # .. add the NAL starting sequence
            self.byte_offset += 1
        else:
            if self.end:
                logger.debug('<<<<< last fragment found')
            else:
                logger.debug('<<<<< intermediate fragment found')
            head = bytes()
            self.byte_offset += 1

        if self.fragment_type not in NALType.supported():
            raise RTPDecodeError(f'Got unsupported NALType {self.fragment_type} for packetization-mode 1')
        elif self.fragment_type == NALType.STAP_A:
            logger.debug('<<<<< STAP_A NALType: (%d), decoding', self.fragment_type)
        elif self.fragment_type == NALType.FU_A:  # This code only handles "Type"  =  28, i.e. "FU-A"
            logger.debug('<<<<< FU_A NALType: (%d), decoding', self.fragment_type)
        return head + self.raw[self.byte_offset:]



"""Deal with SDP body from the RTSP session creation"""

class Media:
    """A media stream"""

    def __init__(self, name, port, profile, payload_type, **kwargs):
        self.name = name
        self.port = port
        self.profile = profile
        self.payload_type = payload_type
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        profile = self.profile.replace('/', '.')
        return '_'.join([self.name, profile, self.payload_type])


class SDP:
    """
    This class represents and SDP Payload.

    Example sdp:
    v=0
    o=- 1 1 IN IP4 127.0.0.1
    s=Test
    a=type:broadcast
    t=0 0
    c=IN IP4 0.0.0.0
    m=video 0 RTP/AVP 96
    a=rtpmap:96 H264/90001
    a=fmtp:96 packetization-mode=1; profile-level-id=4d0029;; sprop-parameter-sets=,
    a=control:track0
    m=audio 0 RTP/AVP 8
    a=control:track1
    """

    @staticmethod
    def parse_attribute(description):
        attrname, value = description.split(':')
        return attrname, value

    def __init__(self, raw):
        self.media = []
        self._raw = raw
        self._lines = self._raw.splitlines()

        new_media = None
        body_lines = self._lines
        body_lines.reverse()
        while body_lines:
            line = body_lines.pop()
            sdp_code, _, description = line.partition('=')
            if sdp_code == 'v':
                self.version = int(description.strip())
            elif sdp_code == 'o':
                self.originator = description
            elif sdp_code == 'a':
                attribute, value = SDP.parse_attribute(description)
                if new_media is None:
                    setattr(self, attribute, value)
                else:
                    setattr(new_media, attribute, value)
            elif sdp_code == 't':
                self.time = tuple(map(int, description.split()))
            elif sdp_code == 's':
                self.name = description.strip()
            elif sdp_code == 'c':
                self.host_info = description
            elif sdp_code == 'm':
                if new_media is not None:
                    self.media.append(new_media)
                new_media = Media(*description.split())
        if new_media:
            self.media.append(new_media)

    def __iter__(self):
        return (m for m in self.media)



"""Exceptions for the mebo robot"""
class MeboError(Exception):
    """General Error for Mebo"""


class MeboCommandError(MeboError):
    """Exception raised when a command is incorrect"""


class MeboRequestError(MeboError):
    """Exception raised when a request is not accepted by Mebo command server"""


class MeboConnectionError(MeboError):
    """Exception raised when connection to mebo fails"""


class MeboConfigurationError(MeboError):
    """Exception raised when a Mebo misconfiguration is detected"""


class MeboDiscoveryError(MeboError):
    """Exception raised when unable to autodiscover mebo on the LAN"""



import os

from hashlib import md5


def _hash(s):
    return md5(s).hexdigest()


def _combine(*args, encoding='utf-8'):
    return ':'.join(args).encode('utf-8')


def ha1(username, realm, password=None):
    if password is None:
        password = os.getenv('STREAM_PASSWORD')
    if not password:
        raise Exception('supply stream password')
    return _hash(_combine(username, realm, password))


def ha2(method, uri):
    return _hash(_combine(method, uri))


def challenge_response(nonce, username, realm, password, method, uri):
    """ Calculates the challenge response for digest authentication. Detials are
    available here: https://tools.ietf.org/html/rfc2617

    :param nonce: The server generated nonce issued in the challenge response
    :param username: the username for to be used for digest authentication
    :param realm: The realm (namespace) of the protected information
    :param password: The password used to protect the information
    :param method: The method used for the request (i.e, GET or DESCRIBE, depending on protocol)
    :param uri: THe uri of the requested resource
    >>> nonce = '595ebd874c33ec0efa0aa306077ae6304c4573c7ad70c0b583298ab2ada7e1a6'
    >>> username = 'stream'
    >>> realm = 'realm'
    >>> method = 'DESCRIBE'
    >>> uri = 'rtsp://172.16.1.107/streamhd/ '
    >>> expected_response = '91b3cd804ec214aeaebfcf51306a093a'
    >>> calculated_response = challenge_response(nonce, username, realm, None, method, uri)
    >>> assert calculated_response == expected_response
    """
    h1 = ha1(username, realm, password)
    h2 = ha2(method, uri)
    return _hash(_combine(h1, nonce, h2))



__title__ = 'mebo'
__description__ = 'Simple python interface to control the mebo toy robot'
__url__ = 'https://github.com/crlane/python-mebo'
__version__ = '0.1.0.b1'
__author__ = 'Cameron Lane'
__author_email__ = 'crlane@adamanteus.com'
__license__ = 'MIT'
__copyright__ = 'Copyright 2018 Cameron Lane'



import re

from collections import namedtuple

StatusLine = namedtuple('StatusLine', ['protocol', 'status_code', 'reason'])
NONCE_MATCHER = re.compile(r'nonce="(?P<digest_nonce>[a-f0-9]{64})"')


class RTSPRequest:

    def __init__(self, method, url, protocol, **headers):
        self.url = url
        self.method = method
        self._headers = headers
        self._command = ' '.join([method, url, protocol])

    @property
    def headers(self):
        if self._headers is None:
            self._headers = {}
        return self._headers

    @property
    def raw_text(self):
        string = '\r\n'.join(self.text())
        return bytes(string.encode('utf-8'))

    def text(self):
        yield self._command
        for header, value in self.headers.items():
            yield f'{header}: {value}'
        yield '\r\n'


class RTSPResponse:
    '''Represents responses for RTSP Negotiation

    Clones much OF the behavior of the requests' HTTPResponse object
    '''

    def __init__(self, request, text, encoding='utf-8'):
        self.request = request
        self._raw = text

        self.encoding = encoding
        self.fallback = 'ascii'

        self._headers = {}
        self._status_line = None

    def __iter__(self):
        return self.iter_lines()

    def iter_lines(self):
        return (l.decode(self.encoding) for l in self._raw.splitlines())

    @property
    def lines(self):
        return list(self.iter_lines())

    @property
    def content(self):
        return self._raw

    @property
    def text(self):
        try:
            return self.content.decode(self.encoding)
        except UnicodeDecodeError:
            return self.content.decode(self.fallback)

    @property
    def nonce(self):
        try:
            digest_challenge_header = self.headers.get('WWW-Authenticate')
            match = NONCE_MATCHER.search(digest_challenge_header)
            if match:
                return match.group('digest_nonce')
        except KeyError:
            print(f'No challenge header found: {self.headers}!')
            raise

    @property
    def _status(self):
        if self._status_line is None:
            self._status_line = StatusLine(*self.lines[0].split())
        return self._status_line

    @property
    def headers(self):
        if not self._headers:
            for line in self.lines[1:]:
                if not line:
                    break
                header, value = re.split(r':\s?', line, maxsplit=1)
                self._headers[header] = value
        return self._headers

    @property
    def reason(self):
        return self._status.reason

    @property
    def status_code(self):
        return int(self._status.status_code)

    @property
    def protocol(self):
        return self._status.protocol

    @property
    def body(self):
        return self.text.rpartition('\r\n\r\n')[-1]


class RTSPMediaError(Exception):
    pass



import logging
import random
import socket
import threading

from urllib.parse import urlsplit, urljoin

from mebo.stream.auth import challenge_response

from mebo.stream.rtp import (
    RTPStream
)

from mebo.stream.sdp import (
    SDP
)


from mebo.stream.rtsp import (
    RTSPResponse,
    RTSPRequest,
)

from mebo.stream import PROTOCOL_VERSION

logger = logging.getLogger(__name__)


class RTSPSession:

    USER_AGENT = 'python-mebo'

    def __init__(self, url, port=None, username=None, realm=None, password=None, user_agent=None):
        self.url = url
        scheme, host, path, _, _ = urlsplit(url)
        self.host = host
        self.port = port or 554  # port 554 is default
        if not user_agent:
            user_agent = RTSPSession.USER_AGENT
        self._user_agent = user_agent

        # Digest auth
        self._username = username
        if not username and realm:
            raise RTSPSessionConfigurationError('Must supply url, username, and realm')

        self._realm = realm
        self._password = password
        self._cnonce = None
        self._nc = 1
        self._opaque = ''
        self._cseq = 1

        # private sockets for rtsp
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._socket.connect((self.host, self.port))

        # rtp info
        self._session_id = None
        self.audio_stream = None
        self.video_stream = None

    def digest_response(self, nonce, method):
        return challenge_response(nonce, self._username, self._realm, self._password, method, self.url)

    def _authorization(self, resp):
        assert resp.nonce
        username = f'Digest username="{self._username}"'
        realm = f'realm="{self._realm}"'
        nonce = f'nonce="{resp.nonce}"'
        uri = f'uri="{self.url}"'
        nc = f'nc={self._nc:0>8}'
        cnonce = f'cnonce="{self._cnonce}"'
        qop = 'qop='
        response = f'response="{self.digest_response(resp.nonce, resp.request.method)}"'
        logger.debug(f'Got authorization response: {response}')
        opaque = 'opaque=""'
        msg = ','.join([username, realm, nonce, uri, nc, cnonce, qop, response, opaque])
        logger.debug(f'Message for challenge generated: {msg}')
        return msg

    def _challenge(self, resp):
        """
        :param resp: The original response which issued the challenge nonce
        """
        if not resp.nonce:
            raise RTSPSessionError(f'Unable to find nonce in challenge header')
        self._cnonce = random.randrange(0, 10**8)
        resp.request.headers.update(**{'Authorization': self._authorization(resp)})
        try:
            return self._request(resp.request, challenge=False)
        except Exception as e:
            raise DigestChallengeError('Unable to complete challenge')
        finally:
            self._nc += 1

    def _default_headers(self):
        return {
            'User-Agent': self._user_agent,
            'CSeq': self._cseq
        }

    def _request_factory(self, method, url, protocol=PROTOCOL_VERSION, **kwargs):
        h = self._default_headers()
        h.update(**kwargs)
        return RTSPRequest(method, url, protocol, **h)

    def _request(self, request, challenge=True):
        logger.debug('Request sent: %s', request.raw_text)
        bytes_sent = self._socket.send(request.raw_text)
        assert bytes_sent
        response = self._socket.recv(4096)
        logger.debug('Response received: %s', response)
        resp = RTSPResponse(request, response)
        if resp.status_code == 200:
            return resp
        elif resp.status_code == 401 and challenge:
            return self._challenge(resp)
        else:
            raise RTSPSessionError('Unable to complete digest challenge')

    def start_streams(self):

        desc = self.describe()
        self._cseq += 1
        logger.debug('SDP: %s', desc.body)
        self.sdp = SDP(desc.body)
        self.streams = [RTPStream(media_desc, host=self.host) for media_desc in self.sdp.media]

        for stream in self.streams:
            extra_headers = {'Transport': stream.transport}
            if self._session_id is not None:
                extra_headers.update({'Session': self._session_id})
            track = self.setup(urljoin(self.url, stream.name), **extra_headers)
            assert track.status_code == 200
            logger.debug('%s response body: %s', stream.name, track.lines)
            track_id = track.headers.get('Session')
            logger.debug('Track session id: %s', track_id)
            if self._session_id is None:
                self._session_id = track_id
            elif self._session_id != track_id:
                logger.warning('Got an extra session id: %s', self._session_id)
            self._cseq += 1
            stream.setup(track.headers.get('Transport'))

        logger.debug('RTP streams established: %s', self.streams)

        if self.streams:
            play_response = self.play(**{'Session': self._session_id, 'Range': 'npt=0.000-'})
            assert play_response.status_code == 200
            self._cseq += 1
            logger.debug('Play response: %s', play_response.lines)

            for stream in self.streams:
                ext = 'h264' if stream.sdp.name == 'video' else 'g711a'
                name = f'{stream.sdp}.{stream.sdp.control}.{ext}'
                t = threading.Thread(name=f'{stream.sdp.name}-thread', target=stream.capture, args=(name,), kwargs=dict(packets=4000))
                t.start()

    def options(self, **kwargs):
        # req = RTSPRequest(self.url, 'OPTIONS', **kwargs)
        # return self._request(req)
        raise NotImplementedError('Mebo RTSP server does not implement this')

    def describe(self, **kwargs):
        req = self._request_factory('DESCRIBE', self.url, **kwargs)
        return self._request(req)

    def setup(self, url, **kwargs):
        req = self._request_factory('SETUP', url, **kwargs)
        return self._request(req)

    def play(self, **kwargs):
        req = self._request_factory('PLAY', self.url, **{'Session': self._session_id, 'Range': 'npt=0.000-'})
        return self._request(req)

    def pause(self, **kwargs):
        raise NotImplementedError

    def teardown(self, **kwargs):
        raise NotImplementedError

    def record(self, **kwargs):
        raise NotImplementedError


class RTSPSessionConfigurationError(Exception):
    pass


class RTSPSessionError(Exception):
    pass


class DigestChallengeError(Exception):
    pass



#!/usr/bin/env python
import os
from setuptools import (
    setup,
    find_packages
)

# NOTE: thanks @kennethreitz for the inspiration here. ;)
here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, 'mebo', '__version__.py'), 'r') as f:
    exec(f.read(), about)


def find_requirements(filename='requirements.txt'):
    with open(filename, 'r') as f:
        return f.readlines()


def readme(filename='README.rst'):
    with open(filename, 'r') as f:
        return f.read()

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme(),
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    package_data={'': ['LICENSE.md']},
    packages=find_packages(exclude=['contrib', 'docs', 'test*']),
    install_requires=find_requirements(),
    license=about['__license__'],
    extras_require={
        'testing': find_requirements('test_requirements.txt'),
        'development': ['ipython', 'ipdb']
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)
